/*
 * tetris.c
 *
 * A Tetris game for OpenRISC. This is pretty much the same as
 * any other Tetris game you have ever played. 
 *
 * You can press 'l' for a level warp cheat. The cheat takes effect
 * when your current block reaches the bottom - so don't press 'l' more
 * than a few times, even if it doesn't seem to be doing anything.
 *
 */


#include <stdlib.h>
#include <string.h>
#include <stdio.h>



typedef short byte_t;
typedef enum { FALSE = 0, TRUE = 1 } BOOL;

#define TICKSCALE	10

#define REPORT          1
#define DEMO_ACTIVE     2

#define TWIDTH		12
#define THEIGHT		24
#define TX			4
#define TY			0
#define SCORE_X		25
#define SCORE_Y		4
#define PREVIEW_X	25
#define PREVIEW_Y	20

#define RECT_BYTE   '/'

/*
 *
 * Interface to the VGA driver.
 *
 */

#define BLOCK_CHAR	1
#define LINE_CHAR	2
#define LINE_ATTR	0x04
#define TYPE_1_ATTR	0x3f
#define TYPE_2_ATTR	0x77
#define TYPE_3_ATTR	0x58
#define TYPE_4_ATTR	0x70
#define TYPE_5_ATTR	0x20
#define BLACK_ATTR	0x00
#define GAME_OVER_ATTR	0x04

//static const Font_Data block = { { 0xff , 0xab , 0xd5 , 0xab ,
//								0xd5 , 0xab , 0xd5 , 0xff } } ;
//static const Font_Data vert_line = { { 0x18 , 0x18 , 0x18 , 0x18 ,
//								0x18 , 0x18 , 0x18 , 0x18 } } ;
//static const Colour_Data line_attr = { 2 , 2 , 2 } ;
//static const Colour_Data type_1 = { 3 , 0 , 0 } ;
//static const Colour_Data type_2 = { 0 , 3 , 0 } ;
//static const Colour_Data type_3 = { 3 , 3 , 0 } ;
//static const Colour_Data type_4 = { 0 , 0 , 3 } ;
//static const Colour_Data type_5 = { 3 , 0 , 3 } ;
//static const Colour_Data black_attr = { 0 , 0 , 0 } ;
//static const Colour_Data game_over_fore_attr = { 3 , 3 , 0 } ;
//static const Colour_Data game_over_back_attr = { 0 , 0 , 3 } ;








#define NUM_PIECES				5
#define NUM_BLOCKS_PER_PIECE	4

/*
 * Game data and game functions
 *
 */
typedef struct Block_Data_struct {
			byte_t		x , y ;
			} Block_Data ;

typedef struct Piece_Data_struct {
			Block_Data	block_list [ NUM_BLOCKS_PER_PIECE ] ;
			byte_t		attr ;
			byte_t		x , y ;
			BOOL		rotatable ;
			BOOL		is_preview ;
			} Piece_Data ;

typedef struct Collision_Table_struct {
			BOOL		ct_matrix [ TWIDTH ][ THEIGHT ] ;
			byte_t		ct_matrix_row_sum [ THEIGHT ] ;
			} Collision_Table ;

typedef enum { NO_COLLISION , OFF_BOARD , HIT_OTHER_BLOCK } Collision_Type ;

const static Piece_Data piece_data_table [ NUM_PIECES ] = {
	{ { { -1 , 0 } ,	/*  abc */
		{ 0 , 0 } ,		/*    d */
		{ 1 , 0 } ,
		{ 1 , 1 } } , TYPE_1_ATTR , 0 , 0 , TRUE , FALSE } ,

	{ { { 0 , 0 } ,		/*  ab */
		{ 1 , 0 } ,		/*  cd */
		{ 0 , 1 } ,
		{ 1 , 1 } } , TYPE_2_ATTR , 0 , 0 , FALSE , FALSE } ,

	{ { { -1 , 0 } ,	/*  abc */
		{ 0 , 0 } ,		/*  d   */
		{ 1 , 0 } ,
		{ -1 , 1 } } , TYPE_3_ATTR , 0 , 0 , TRUE , FALSE } ,

	{ { { -1 , 0 } ,	/*  abc */
		{ 0 , 0 } ,		/*   d  */
		{ 1 , 0 } ,
		{ 0 , 1 } } , TYPE_4_ATTR , 0 , 0 , TRUE , FALSE } ,

	{ { { -1 , 0 } ,	/*  abcd */
		{ 0 , 0 } ,		/*       */
		{ 1 , 0 } ,
		{ 2 , 0 } } , TYPE_5_ATTR , 0 , 0 , TRUE , FALSE } } ;

#include "../stage2boot/gfxlib.c"

static unsigned _timer_base = 0;
static unsigned _timer_zero_at = 0;
static unsigned _previous_timer = 0;
static unsigned char _shadow_text[40*25];
static unsigned char _shadow_attr[40*25];
static volatile unsigned * buttons;

static unsigned get_timer ()
{
    unsigned t = vga [ 0 ] & 0xff;
    if (t < _previous_timer) {
        _timer_base += 0x100;
    }
    _previous_timer = t;
    t += _timer_base;
    return t;
}

static void timer_setup ( unsigned countdown, void * ignore)
{
    _timer_zero_at = get_timer () + countdown ;
}

static unsigned timer_get_ticks ()
{
    unsigned t = get_timer () ;
    if (t >= _timer_zero_at) {
        return 0;
    } else {
        return _timer_zero_at - t;
    }
}


static void vga_write_char ( unsigned x , unsigned y , 
    unsigned char byte, unsigned char colour )
{
    _shadow_text[ x + (y * 40) ] = byte;
    _shadow_attr[ x + (y * 40) ] = colour;
    if (byte == RECT_BYTE) {
        unsigned l, c, i;
        x *= 8;
        y *= 8;
        l = (x / 4) + (y * 80);
        c = colour;
        c |= c << 8;
        c |= c << 16;
        for (i = 0; i < 8; i ++) {
            ((unsigned *) video) [ l ] = c;
            ((unsigned *) video) [ l + 1 ] = c;
            l += 80;
        }
    } else {
        Text_Char(x * 8, y * 8, byte, colour);
    }
}

static void vga_write_string ( unsigned x , unsigned y , 
    const char * s, unsigned char colour )
{
    size_t size = strlen(s);
    strncpy((char *) &_shadow_text[ x + (y * 40) ], s, size);
    memset(&_shadow_attr[ x + (y * 40) ], colour, size);
    Text(x * 8, y * 8, s, colour);
}

static unsigned char vga_read_char ( unsigned x , unsigned y ) 
{
    return _shadow_text[ x + (y * 40) ];
}

static unsigned char vga_read_attrib ( unsigned x , unsigned y ) 
{
    return _shadow_attr[ x + (y * 40) ];
}

static void vga_clear ()
{
    memset(video, 0, 64000);
    memset(_shadow_text, 0, 40*25);
    memset(_shadow_attr, 0, 40*25);
}

static void rotate_piece (  Piece_Data *	p ,
							byte_t			rotate_by )
{
	byte_t		i , x , y ;

	if ( ! p -> rotatable )
	{
		return ;
	}
	for ( ; rotate_by > 0 ; rotate_by -- )
	{
		for ( i = 0 ; i < NUM_BLOCKS_PER_PIECE ; i ++ )
		{
			x = p -> block_list [ i ] . x ;
			y = p -> block_list [ i ] . y ;

			p -> block_list [ i ] . x = 0 - y ;
			p -> block_list [ i ] . y = x ;
		}
	}
}


static Collision_Type collision ( Piece_Data * piece , Collision_Table * ct )
{
	byte_t		i , x , y ;

	for ( i = 0 ; i < NUM_BLOCKS_PER_PIECE ; i ++ )
	{
		x = ( piece -> block_list [ i ] . x ) + ( piece -> x ) ;
		y = ( piece -> block_list [ i ] . y ) + ( piece -> y ) ;

		if (( x < 0 )
		|| ( x >= TWIDTH )
		|| ( y < 0 )
		|| ( y >= THEIGHT ))
		{
			return OFF_BOARD ; /* This is off the board */
		}
		
		if ( ct -> ct_matrix [ x ][ y ] )
		{
			return HIT_OTHER_BLOCK ; /* This is a collision */
		}
	}
	return NO_COLLISION ;
}

static unsigned demo_data_ptr = 0;

#include "demo.c"

static unsigned char get_demo_byte ()
{
    unsigned char ch;

    if (demo_data_ptr >= demo_len) {
        return '?';
    }
    ch = demo_data[ demo_data_ptr ];
    demo_data_ptr ++;
    return ch;
}

static int demo_rand ( int demo_flags ) 
{
    int r;

    if (demo_flags & DEMO_ACTIVE) {
        unsigned i, v = 0;
        unsigned char ch = get_demo_byte();
        if (ch != '#') {
            print ("Missing marker: ");
            printhex (ch);
            print (" @ ");
            printhex (demo_data_ptr);
            exit (1);
        }
        for (i = 0; i < 8; i ++) {
            ch = get_demo_byte();
            v = v << 4;
            if (('0' <= ch) && (ch <= '9')) {
                v |= (unsigned) (ch - '0');
            } else if (('a' <= ch) && (ch <= 'f')) {
                v |= (unsigned) (ch + 10 - 'a');
            } else {
                print ("Not a hex digit: ");
                printhex (ch);
                exit (1);
            }
        }
        r = (int) v;
    } else {
        r = rand();
    }
    if (demo_flags & REPORT) {
        print("#");
        printhex(r);
    }
    return r;
}

static void create_preview_piece ( Piece_Data * new_piece, int demo_flags )
{
	/* Choose a piece and rotate it 0 to 3 times */
	memcpy ( new_piece , & piece_data_table [ 
                demo_rand ( demo_flags ) % NUM_PIECES ] , 
												sizeof ( Piece_Data ) ) ;

	rotate_piece ( new_piece , demo_rand ( demo_flags ) % 4 ) ;

	new_piece -> is_preview = TRUE ;
}

static BOOL add_piece_to_board ( Piece_Data * preview_piece ,
						Piece_Data * new_piece , Collision_Table * ct )
{
	memcpy ( new_piece , preview_piece , sizeof ( Piece_Data ) ) ;

	new_piece -> is_preview = FALSE ;
	
	/* Set up the location */
	new_piece -> x = ( TWIDTH / 2 ) ;
	new_piece -> y = 0 ;

	/* Does the new piece collide with anything? */
	if ( collision ( new_piece , ct ) == NO_COLLISION )
	{
		return TRUE ; /* no */
	}

	/* Yes. Hmm.. well, try moving it down by one or two rows. */
	for ( new_piece -> y = 0 ; new_piece -> y <= 2 ; new_piece -> y ++ )
	{
		if ( collision ( new_piece , ct ) == NO_COLLISION )
		{
			return TRUE ; /* No collision - good */
		}
	}
	/* It still collides in those positions, so the board must be full. */

	return FALSE ;
}


static void prepare_board_in_ct ( Collision_Table * ct )
{
	/* Initialise the collision table */
	byte_t	x , y ;

	for ( y = 0 ; y < THEIGHT ; y ++ )
	{
		for ( x = 0 ; x < TWIDTH ; x ++ )
		{
			ct -> ct_matrix [ x ][ y ] = FALSE ;
		}
		ct -> ct_matrix_row_sum [ y ] = 0 ;
	}
}

static void remove_line_in_ct ( byte_t line , Collision_Table * ct )
{
	byte_t	x , y , yi ;

	/* Remove the line in the collision table */
	/* Everything above it moves down, so all lines from 0..line-1 move
	 * to 1..line */
	for ( y = line ; y >= 0 ; y -- )
	{
		yi = y - 1 ;

		if ( yi >= 0 )
		{
			for ( x = 0 ; x < TWIDTH ; x ++ )
			{
				ct -> ct_matrix [ x ][ y ] = ct -> ct_matrix [ x ][ yi ] ;
			}
			ct -> ct_matrix_row_sum [ y ] = ct -> ct_matrix_row_sum [ yi ] ;
		} else {
			for ( x = 0 ; x < TWIDTH ; x ++ )
			{
				ct -> ct_matrix [ x ][ y ] = FALSE ;
			}
			ct -> ct_matrix_row_sum [ y ] = 0 ;
		}
	}
}


static BOOL user_move_piece ( Piece_Data * piece , 
					byte_t horiz , byte_t vert , Collision_Table * ct )
{
	/* We'll try applying this move and see if a collision occurs. */
	piece -> x += horiz ;
	piece -> y += vert ;
	if ( collision ( piece , ct ) == NO_COLLISION )
	{
		return TRUE ; /* No collision - good */
	}
	/* Unapply the move */
	piece -> x -= horiz ;
	piece -> y -= vert ;
	return FALSE ;
}

static BOOL attempt_rotation ( Piece_Data * copy_of_piece ,
						Piece_Data * piece ,
						Collision_Table * ct ,
						byte_t fudge_factor )
{
	Collision_Type	col_type ;

	/* Attempt the rotation on a copy of the piece */
	memcpy ( copy_of_piece , piece , sizeof ( Piece_Data ) ) ;

	/* Apply the fudge factor - the piece is moved to one side
	 * by this factor so that you can rotate a piece on the edge of
	 * the board. */
	if ( piece -> x > ( TWIDTH / 2 ))
	{
		copy_of_piece -> x -= fudge_factor ; /* move left */
	} else {
		copy_of_piece -> x += fudge_factor ; /* move right */
	}


	rotate_piece ( copy_of_piece , 1 ) ;
	col_type = collision ( copy_of_piece , ct ) ;
	
	if ( col_type == HIT_OTHER_BLOCK )
	{
		/* There was a collision with another block,
		 * so this rotation can't be done */
		return FALSE ;
	} else if ( col_type == OFF_BOARD )
	{
		/* Perhaps the rotation can be done with a larger fudge
		 * factor. We will try fudge factors up to 2 */
		if ( fudge_factor < 2 )
		{
			return attempt_rotation ( copy_of_piece , 
							piece , ct , fudge_factor + 1 ) ;
		} else {
			return FALSE ;
		}
	} else {
		/* No collision, apply the rotation to the working copy */
		memcpy ( piece , copy_of_piece , sizeof ( Piece_Data ) ) ;
		return TRUE ;
	}
}

static BOOL user_rotate_piece ( Piece_Data * piece , Collision_Table * ct )
{
	Piece_Data		copy_of_piece ;

	return attempt_rotation ( & copy_of_piece , piece , ct , 0 ) ;
}



static void draw_piece ( Piece_Data * piece , BOOL positive_drawing )
{
	byte_t		i , x , y ;

	for ( i = 0 ; i < NUM_BLOCKS_PER_PIECE ; i ++ )
	{
		if ( piece -> is_preview )
		{
			x = ( piece -> block_list [ i ] . x ) + PREVIEW_X ;
			y = ( piece -> block_list [ i ] . y ) + PREVIEW_Y ;
		} else {
			x = ( piece -> block_list [ i ] . x ) + ( piece -> x ) + TX + 1 ;
			y = ( piece -> block_list [ i ] . y ) + ( piece -> y ) + TY ;
		}

		if ( positive_drawing )
		{
			vga_write_char ( x , y , RECT_BYTE, piece -> attr ) ;
		} else {
			vga_write_char ( x , y , ' ' , BLACK_ATTR ) ;
		}
	}
}

void prepare_board_in_vga ( void )
{
	int		y, i ;

	vga_clear () ;

	//vga_set_font_character ( BLOCK_CHAR , & block ) ;
	//vga_set_font_character ( LINE_CHAR , & vert_line ) ;
	//vga_set_palette ( LINE_ATTR , 0 , 1 , & line_attr ) ;
	//vga_set_palette ( TYPE_1_ATTR , 0 , 1 , & type_1 ) ;
	//vga_set_palette ( TYPE_2_ATTR , 0 , 1 , & type_2 ) ;
	//vga_set_palette ( TYPE_3_ATTR , 0 , 1 , & type_3 ) ;
	//vga_set_palette ( TYPE_4_ATTR , 0 , 1 , & type_4 ) ;
	//vga_set_palette ( TYPE_5_ATTR , 0 , 1 , & type_5 ) ;
	//vga_set_palette ( BLACK_ATTR , 0 , 1 , & black_attr ) ;
	//vga_set_palette ( GAME_OVER_ATTR , 0 , 1 , & game_over_fore_attr ) ;
	//vga_set_palette ( GAME_OVER_ATTR , 0 , 0 , & game_over_back_attr ) ;

//	for ( y = TY ; y < ( THEIGHT + TY ) ; y ++ )
//	{
//		vga_write_char ( TX , y , LINE_CHAR , LINE_ATTR ) ;
//		vga_write_char ( TX + TWIDTH + 1 , y , LINE_CHAR , LINE_ATTR ) ;
//	}
    for ( y = 0 ; y < (8*24); y ++) {
        for ( i = 0; i < 4 ; i ++) {
            video [ ( y * 320 ) + ( TX * 8 ) + 2 + i ] = 0x4;
            video [ ( y * 320 ) + (( TX + TWIDTH + 1 ) * 8 ) + 2 + i ] = 0x4;
        }
    }

	vga_write_string ( SCORE_X , SCORE_Y , "Lines" , LINE_ATTR ) ;
	vga_write_string ( SCORE_X , SCORE_Y + 3 , "Score" , LINE_ATTR ) ;
	vga_write_string ( SCORE_X , SCORE_Y + 6 , "Level" , LINE_ATTR ) ;
	vga_write_string ( PREVIEW_X , PREVIEW_Y - 3 , "Next" , LINE_ATTR ) ;
}

static void remove_line_in_vga ( byte_t line )
{
	/* This is much the same as remove_line_in_ct, except that
	 * the table is the video memory. We could do this much faster if
	 * we directly manipulated video RAM but that would be a poor example
	 * of using the VGA driver (vga.c) */
	
	byte_t	x , y , yi ;

	/* Everything above it moves down, so all lines from 0..line-1 move
	 * to 1..line */
	for ( y = ( line + TY ) ; y >= TY ; y -- )
	{
		yi = y - 1 ;

		if ( yi >= TY )
		{
			for ( x = TX + 1 ; x < ( TWIDTH + TX + 1 ) ; x ++ )
			{
				vga_write_char ( x , y ,
						vga_read_char ( x , yi ) , 
						vga_read_attrib ( x , yi ) ) ;
			}
		} else {
			for ( x = TX + 1 ; x < ( TWIDTH + TX + 1 ) ; x ++ )
			{
				vga_write_char ( x , y , 0 , BLACK_ATTR ) ;
			}
		}
	}
}

static byte_t freeze_piece ( Piece_Data * piece , Collision_Table * ct )
{
	byte_t		i , x , y ;
	byte_t		complete_lines = 0 ;

	/* Freeze it in the collision table */
	for ( i = 0 ; i < NUM_BLOCKS_PER_PIECE ; i ++ )
	{
		x = ( piece -> block_list [ i ] . x ) + ( piece -> x ) ;
		y = ( piece -> block_list [ i ] . y ) + ( piece -> y ) ;

		ct -> ct_matrix [ x ][ y ] = TRUE ;
		ct -> ct_matrix_row_sum [ y ] ++ ;
	}

	/* Ensure the piece has been drawn. */
	draw_piece ( piece , TRUE ) ;

	/* Now each row may be a complete line. If it is, then
	 * the row sum will be equal to TWIDTH. */
	for ( y = 0 ; y < THEIGHT ; y ++ )
	{
		if ( ct -> ct_matrix_row_sum [ y ] >= TWIDTH )
		{
			/* This is a complete line. Remove it. */
			remove_line_in_ct ( y , ct ) ;
			remove_line_in_vga ( y ) ;
			complete_lines ++ ;
		}
	}
	return complete_lines ;
}



void tetris_main ( int demo_flags )
{
	Collision_Table		ct ;
	Piece_Data			piece ;
	Piece_Data			preview_piece ;
	BOOL				in_game , falling , quit , timeout , event ;
	int					num_ticks, ch ;
	unsigned			rows_fallen = 0 ;
	unsigned			lines = 0 ;
	unsigned			score = 0 ;
	unsigned			level = 1 ;
	unsigned			level_warp = 0 ;
	char				tmp_str [ 10 ] ;
    unsigned            buttons_then [ 8 ];

    if (demo_flags & REPORT) 
        print("\r\n\r\nstart\r\n");

    demo_data_ptr = 0;

	prepare_board_in_vga () ;
	prepare_board_in_ct ( & ct ) ;

	create_preview_piece ( & preview_piece, demo_flags ) ;
    Recache_Buttons();
    memcpy (buttons_then, (unsigned *) buttons, 8 * 4);


	in_game = TRUE ;
	falling = FALSE ;
	quit = FALSE ;
	while ( in_game )
	{
		/* Update the score, etc */
		sprintf ( tmp_str , "%5u" , lines ) ;
		vga_write_string ( SCORE_X , SCORE_Y + 1 , tmp_str , LINE_ATTR ) ;

		sprintf ( tmp_str , "%5u" , score ) ;
		vga_write_string ( SCORE_X , SCORE_Y + 4 , tmp_str , LINE_ATTR ) ;

		sprintf ( tmp_str , "%5u" , level ) ;
		vga_write_string ( SCORE_X , SCORE_Y + 7 , tmp_str , LINE_ATTR ) ;

		draw_piece ( & preview_piece , FALSE ) ;

		/* Now start a piece falling */
		if ( add_piece_to_board ( & preview_piece , & piece , & ct ) )
		{
			/* new piece created ok */
			falling = TRUE ;
			rows_fallen = 0 ;

			/* create a new preview piece */
			create_preview_piece ( & preview_piece, demo_flags ) ;
			draw_piece ( & preview_piece , TRUE ) ;
		} else {
			in_game = FALSE ;
		}

		/* Each tick is 5 milliseconds.
		 * We initially start with the blocks falling at one row per second
		 * (i.e. every 200 ticks), then increase by 20 ticks (10%) after 5 lines
		 * to a limit of one row per 100ms (every 20 ticks)
		 */
		num_ticks = ( 220 - ( 20 * (int) level ) ) ;

		if ( num_ticks < 15 )
		{
			num_ticks = 15 ;
		}

		timer_setup ( num_ticks / TICKSCALE , NULL ) ;
		while ( falling )
		{
			draw_piece ( & piece , TRUE ) ;

			timeout = FALSE ;
            event = FALSE ;

			/* wait for key press or the timer reaching 0 */
            while ((!(demo_flags & DEMO_ACTIVE))
            && (timer_get_ticks () != 0) && !event) {
                Recache_Buttons();
                while (buttons[ BTN_EAST ] > buttons_then[ BTN_EAST ]) {
                    draw_piece ( & piece , FALSE ) ;
                    user_move_piece ( & piece , 1 , 0 , & ct ) ;/* east */
                    draw_piece ( & piece , TRUE ) ;
                    if (demo_flags & REPORT) print("e");
                    buttons_then[ BTN_EAST ] ++;
                    event = TRUE;
                }
                while (buttons[ BTN_WEST ] > buttons_then[ BTN_WEST ]) {
                    draw_piece ( & piece , FALSE ) ;
                    user_move_piece ( & piece , -1 , 0 , & ct ) ;/* west */
                    draw_piece ( & piece , TRUE ) ;
                    if (demo_flags & REPORT) print("w");
                    buttons_then[ BTN_WEST ] ++;
                    event = TRUE;
                }
                while (buttons[ BTN_SOUTH ] > buttons_then[ BTN_SOUTH ]) {
                    draw_piece ( & piece , FALSE ) ;
                    user_move_piece ( & piece , 0 , 1 , & ct ) ;/* south */
                    draw_piece ( & piece , TRUE ) ;
                    if (demo_flags & REPORT) print("s");
                    buttons_then[ BTN_SOUTH ] ++;
                    event = TRUE;
                }
                while (buttons[ ROT_CCW ] > buttons_then[ ROT_CCW ]) {
                    draw_piece ( & piece , FALSE ) ;
                    user_rotate_piece ( & piece , & ct ) ;  /* ccw */
                    user_rotate_piece ( & piece , & ct ) ;
                    user_rotate_piece ( & piece , & ct ) ;
                    draw_piece ( & piece , TRUE ) ;
                    if (demo_flags & REPORT) print("<");
                    buttons_then[ ROT_CCW ] ++;
                    event = TRUE;
                }
                while (buttons[ ROT_CW ] > buttons_then[ ROT_CW ]) {
                    draw_piece ( & piece , FALSE ) ;
                    user_rotate_piece ( & piece , & ct ) ;  /* cw */
                    draw_piece ( & piece , TRUE ) ;
                    if (demo_flags & REPORT) print(">");
                    buttons_then[ ROT_CW ] ++;
                    event = TRUE;
                }
                while (buttons[ BTN_CENTRE ] > buttons_then[ BTN_CENTRE ]) {
                    draw_piece ( & piece , FALSE ) ;
                    if (!timeout) {
                        while ( user_move_piece ( & piece , 0 , 1 , & ct ) ) {}
                        timeout = TRUE;
                    }
                    draw_piece ( & piece , TRUE ) ;
                    if (demo_flags & REPORT) print("d");
                    buttons_then[ BTN_CENTRE ] ++;
                    event = TRUE;
                }
                if (!(buttons[ SWITCHES ] & 1)) {
                    in_game = FALSE ;
                    falling = FALSE ;
                    quit = TRUE ;
                    event = TRUE;
                }
            }

            ch = 0;
            while ((demo_flags & DEMO_ACTIVE) 
            && (!timeout) && (ch != 'x')) {
                unsigned v = get_timer();
                draw_piece ( & piece , FALSE ) ;
                ch = get_demo_byte();
                switch (ch) {
                    case 'e' :  user_move_piece ( & piece , 1 , 0 , & ct ) ;
                                event = TRUE;
                                break;
                    case 'w' :  user_move_piece ( & piece , -1 , 0 , & ct ) ;
                                event = TRUE;
                                break;
                    case 's' :  user_move_piece ( & piece , 0 , 1 , & ct ) ;
                                event = TRUE;
                                break;
                    case '<' :  user_rotate_piece ( & piece , & ct ) ;
                                user_rotate_piece ( & piece , & ct ) ;
                                user_rotate_piece ( & piece , & ct ) ;
                                event = TRUE;
                                break;
                    case '>' :  user_rotate_piece ( & piece , & ct ) ;
                                event = TRUE;
                                break;
                    case 'd' :  if (!timeout) {
                                    while ( user_move_piece ( 
                                        & piece , 0 , 1 , & ct ) ) {}
                                }
                                event = TRUE;
                                break;
                    case 't' :  timeout = TRUE;
                                break;
                    case 'x' :  break;
                    default :   print ("Out of sequence byte: ");
                                printhex (ch);
                                exit (1);
                                break;
                }
                draw_piece ( & piece , TRUE ) ;
                v += 10;
                while ((v < get_timer())
                && (timer_get_ticks() != 0)) {}
            }

            while ((demo_flags & DEMO_ACTIVE) 
            && (timer_get_ticks () != 0)) {
                if (!(buttons[ SWITCHES ] & 0x10)) {
                    /* stop demo playback */
                    in_game = FALSE ;
                    falling = FALSE ;
                    quit = TRUE ;
                    event = TRUE;
                }
            }

            if (!event) {
                timeout = TRUE;
            }
            if (timeout) {
                if (demo_flags & REPORT) print("t");
            } else {
                if (demo_flags & REPORT) print("x");
            }

			draw_piece ( & piece , FALSE ) ;

			if (timeout) {
				/* timeout occurred */

				timer_setup ( num_ticks / TICKSCALE , NULL ) ; /* reset timer */
				if ( user_move_piece ( & piece , 0 , 1 , & ct ) )
				{
					rows_fallen ++ ;
				} else {
					/* can't move piece down any more.. freeze it */
					byte_t		n_lines = freeze_piece ( & piece , & ct ) ;

					lines += n_lines ;
					level = ( lines / 5 ) + 1 + level_warp ;

					/* update the score. You get more points when you
					 * can remove several lines at a time: */
					switch ( n_lines )
					{
						case 1 :	score += 10 * level ;
									break ;
						case 2 :	score += 20 * level ;
									break ;
						case 3 :	score += 40 * level ;
									break ;
						case 4 :	score += 80 * level ;
									break ;
						default :	break ;
					}

					/* and you get a bonus for landing a piece quickly */
					if ( rows_fallen < ( THEIGHT / 2 ))
					{
						score += ((( THEIGHT / 2 ) - 
								rows_fallen ) / 4 ) * level ;
					}
					falling = FALSE ;
				}
			}
		} /* while falling */
	} /* while in_game */

    if (demo_flags & DEMO_ACTIVE) {
        return;
    }
	if (!quit) {
		/* The game has finished. Give user a chance to look at the board. */
		vga_write_string ( 0 , ( THEIGHT / 2 ) + TY ,
		/*   0123456789012345678901234567890123456789 */
			"  GAME OVER : Turn off switch 1 now " , GAME_OVER_ATTR ) ;

        while ((buttons[ SWITCHES ] & 1)) { Recache_Buttons(); }
	}
} /* void tetris_main */



void centre_text ( int y , const char * str , int colour )
{
	vga_write_string ( ( 40 - strlen ( str )) / 2 , y , str , colour ) ;
}

int main ( void )
{
    /* initialisation functions: */
    //uart_init () ;
    //vga_init () ;
    //kbd_init () ;
    //timer_init () ;
    //int_enable () ;

    /* main loop: */
    buttons = shm;
	while ( 1 )
	{
		vga_clear () ;

		centre_text ( 12 , "Tetris Game" , 0xb0 ) ;
		centre_text ( 2 , "Department of Computer Science" , 0x70 ) ;
		centre_text ( 4 , "Field Programmable Gate Array Demo" , 0xa0 ) ;

		vga_write_string ( 0 , 23 ,
				"please turn off switches" , 0xfb ) ;

        while ((buttons[ SWITCHES ] & 0x1111)) { Recache_Buttons(); }

		vga_write_string ( 0 , 22 ,
				"turn on switch 1 to play or" , 0xfb ) ;
		vga_write_string ( 0 , 23 ,
				"switch 2 for self playing demo" , 0xfb ) ;

        while (!(buttons[ SWITCHES ] & 0x11)) { Recache_Buttons(); }

        if (buttons[ SWITCHES ] & 0x10) {
            while (buttons[ SWITCHES ] & 0x10) {
                tetris_main ( DEMO_ACTIVE ) ;
            }
        } else if (buttons[ SWITCHES ] & 0x100) {
            tetris_main ( REPORT ) ;
        } else {
            tetris_main ( 0 ) ;
        }
	}
	return 0 ;
}






